(* 多相型関数をインスタンス化する *)

open Type
open Id
open ILang

module M3 = MyMap.Make
  (struct
     type t = Type.t list
     let compare = Pervasives.compare
   end)

(* 型 *)
let typeof x =
  match x with
  | Unit -> Tunit
  | Int _ -> Tint
  | Float _ -> Tfloat
  | Label l -> failwith "typeof"
  | Var x -> match x.typ with
    | Tvar{contents = None} -> failwith
        ("cannot determine type of " ^ Id.to_string x)
    | t -> t

(* 多相型関数 -> 型 -> 型を限定した関数 *)
let instantiate (x,ys,e) ts =
  let tx = Typing.instantiate x.typ in
  match tx with
  | Tfun(ts_, t) -> begin
      List.iter2 Typing.unify ts ts_;
      let tx = Typing.deref_type tx in
      match tx with
      | Tfun(ts, t) ->
          let x_ = new_id_t x.name tx in
          let ys_ = List.map2 new_id_t (List.map (fun y -> y.name) ys) ts in
          let env = List.fold_left2
            (fun m y y_ -> M.add y (Var y_) m)
            (M.add x (Var x_) M.empty) ys ys_
          in
          (x_, ys_, Alpha.g env e)
      | _ -> assert false
    end
  | _ -> failwith "Expansion.instantiate"

(* 多相型関数と型から、すでに生成された関数への写像 *)
let env = ref M.empty

(* 使用する箇所を置き換えたり LetRec を追加したり *)
let rec g funs e =
  let find x ts env = M3.find ts (M.find x env) in
  let add x ts y env = M.add x
    (M3.add ts y (try M.find x env with Not_found -> M3.empty))
    env
  in
  match e with
  | Op(_, xs)
    when List.exists (function Var x -> M.mem x funs | _ -> false) xs ->
      (* この辺はexpansionじゃ無理だろ *)
      failwith "polymorphic closure not supported"
  | V _ | Op(Label _, _) -> e
  | Op(Var x, ys) when not (M.mem x funs) -> e
  | Op(Var x, ys) ->
      let ts = List.map typeof ys in
      let (z,_,_) = try find x ts !env with Not_found ->
        let name,args,body = instantiate (M.find x funs) ts in
        env := add x ts (name,[],V Unit) !env;
        let name,args,body as f = name,args, g funs body in
        env := add x ts f !env; f
      in
      Op(Var z, ys)
  | Op _ -> assert false
  | If(c,x,y,e1,e2) -> If(c,x,y, g funs e1, g funs e2)
  | Let(x,e1,e2) -> Let(x, g funs e1, g funs e2)
  | LetRec(x,ys,e1,e2) ->
      let e_ = match x.typ with
        | ForAll _ -> LetRec(x,ys,e1, g (M.add x (x,ys,e1) funs) e2)
        | _ -> LetRec(x,ys, g funs e1, g funs e2)
      in
      try M3.fold
        (fun _ (x,ys,e1) e2 -> LetRec(x,ys,e1,e2))
        (M.find x !env) e_
      with Not_found -> e_

(* エントリーポイント *)
let f e =
  env := M.empty;
  let ret = g M.empty e in
  env := M.empty; (* メモリ節約 :-p *)
  ret
