(* 覗き穴最適化 *)

open TreeAsm
open LinearAsm

module H = Hashtbl.Make(struct
  type t = string
  let hash = Hashtbl.hash
  let equal x y = String.compare x y = 0
end)

let all_nodes = H.create 100
let funs = H.create 100

let rec used_labels s e =
  if e == end_node then s
  else let s = match e.exp with
  | JUMP x | JAL x | BE(_,_,x) | BL(_,_,x) | BGE(_,_,x)
  | BEI(_,_,x) | BLI(_,_,x) | BGEI(_,_,x)
  | FBL(_,_,x) | FBGE(_,_,x) -> S2.add x s
  | _ -> s
  in
  used_labels s e.cont

let rec elim_labels s e =
  if e == end_node then ()
  else begin
    if not (S2.mem e.label s) then e.label <- nolabel;
    elim_labels s e.cont
  end

let clean xs =
  let s = List.fold_left (fun s x -> used_labels s x.entry) S2.empty xs in
  List.iter
    (fun x -> elim_labels s x.entry;
      x.fbody.label <- x.fname ^ ".body")
    xs

let rec h e =
  if e == end_node then ()
  else begin
    if e.label != nolabel then H.add all_nodes e.label e;
    h e.cont
  end

let modified = ref false

let rec is_tail self e =
  assert (e != end_node);
  e == self.exit ||
  match e with
  | { exp = JUMP x } when H.mem all_nodes x ->
      begin match H.find all_nodes x with
      | e_ when e_ == self.exit -> true
      | { exp = JUMP y } -> e.exp <- JUMP y; is_tail self e
      | _ -> false end
  | _ -> false

let elim e =
  while e.cont.label == nolabel do
    modified := true;
    e.cont <- e.cont.cont;
  done

let rec g self e =
  if e == end_node then ()
  else match e with
  | { exp = JAL x; cont = e_ } when is_tail self e_ -> (* 末尾呼び出し *)
      modified := true;
      elim e;
      begin try
        let func = H.find funs x in
        if func.stk = self.stk then (* スタックフレームを放置できる *)
          e.exp <- JUMP func.fbody.label
        else raise Not_found (* トリック *)
      with Not_found ->
        if self.stk = 0 then e.exp <- JUMP x
        else begin
          e.exp <- LOAD(reg_ra, reg_sp, "0");
          e.cont <- make (ADDI(reg_sp, reg_sp, string_of_int self.stk))
              (make (JUMP x) e.cont)
        end
      end;
      g self e.cont
  | { exp = JUMP x } ->
      elim e;
      let rec iter x =
        try match H.find all_nodes x with
        | { exp = JUMP y } -> modified := true; e.exp <- JUMP y; iter y
        | { exp = JAL y } -> modified := true; e.exp <- JAL y
        | { exp = JR r } -> modified := true; e.exp <- JR r
        | e1 when e1 == e.cont ->
            modified := true;
            e.label <- e1.label;
            e.exp <- e1.exp;
            e.cont <- e1.cont
        | e1 when e1 == self.exit ->
            modified := true;
            let rec copy e2 =
              if e2 == end_node then e.cont
              else make e2.exp (copy e2.cont)
            in
            e.exp <- e1.exp;
            e.cont <- copy e1.cont
        | _ -> ()
        with Not_found -> ()
      in
      iter x;
      g self e.cont
  | { exp = JR x } ->
      elim e;
      g self e.cont
  | _ -> g self e.cont

let f xs =
  Format.eprintf "peephole optimization...@.";
  H.clear funs;
  H.clear all_nodes;
  modified := true;
  while !modified do
    modified := false;
    clean xs;
    List.iter (fun x -> H.add funs x.fname x; h x.entry) xs;
    List.iter (fun x -> g x x.entry) xs;
  done;
  H.clear funs;
  H.clear all_nodes;
  xs
