(* レジスタ割り当て *)

open TreeAsm

let seq (exp, e) = concat (Ret exp) reg_dummy e

let rec alloc cont regenv x =
  (* allocate a register or fail *)
  assert (not (M2.mem x regenv));
  if is_reg x then x else
  let free = fv cont in
  try
    let live = (* 生きているレジスタ *)
      S2.fold
        (fun y live ->
	  if is_reg y then S2.add y live else
          try S2.add (M2.find y regenv) live
          with Not_found -> live)
        free
        S2.empty
    in
    let r = (* そうでないレジスタを探す *)
      List.find
        (fun r -> not (S2.mem r live))
        regs_lst in
    (* Format.eprintf "allocated %s to %s@." x r; *)
    r
  with Not_found ->
    failwith (Printf.sprintf "register allocation failed for %s" x)

(* auxiliary function for g and g'_and_restore *)
let add x r regenv =
  if is_reg x then (assert (x = r); regenv) else
  M2.add x r regenv

(* auxiliary functions for g' *)
exception NoReg of string
let mem x regenv =
  if is_reg x then true
  else M2.mem x regenv
let find x regenv =
  if is_reg x then x else
  try M2.find x regenv
  with Not_found -> raise (NoReg x)

(* 命令列のレジスタ割り当て (caml2html: regalloc_g) *)
let rec g dest cont regenv = function
  | Ret exp -> g'_and_restore dest cont regenv exp
  | Let(x, exp, e) ->
      assert (not (M2.mem x regenv));
      let cont' = concat e dest cont in
      let (e1', regenv1) = g'_and_restore x cont' regenv exp in
      let r = alloc cont' regenv1 x in
      let (e2', regenv2) = g dest cont (add x r regenv1) e in
      (concat e1' r e2', regenv2)
  | Forget(x, e) -> assert false
(* 使用される変数をスタックからレジスタへRestore *)
and g'_and_restore dest cont regenv exp =
  try g' dest cont regenv exp
  with NoReg x ->
    (* Format.eprintf "restoring %s@." x; *)
    g dest cont regenv (Let(x, Restore x, Ret exp))
(* 各命令のレジスタ割り当て *)
and g' dest cont regenv = function
  | Nop | Imm _ | Rd | Restore _ as exp -> (Ret exp, regenv)
  | Mov x -> (Ret(Mov(find x regenv)), regenv)
  | St(x, y, i) -> (Ret(St(find x regenv, find y regenv, i)), regenv)
  | If(c, x, y, e1, e2) as exp -> g'_if dest cont regenv exp
        (fun e1' e2' -> If(c, find x regenv, find y regenv, e1', e2')) e1 e2
  | IfI(c, x, i, e1, e2) as exp -> g'_if dest cont regenv exp
        (fun e1' e2' -> IfI(c, find x regenv, i, e1', e2')) e1 e2
  | FIf(c, x, y, e1, e2) as exp -> g'_if dest cont regenv exp
        (fun e1' e2' -> FIf(c, find x regenv, find y regenv, e1', e2')) e1 e2
  | Op2(o,x,y) -> (Ret(Op2(o, find x regenv, find y regenv)), regenv)
  | Op1i(o,x,i) -> (Ret(Op1i(o, find x regenv, i)), regenv)
  | Op1(o,x) -> (Ret(Op1(o, find x regenv)), regenv)
  | Call(l, ys) as exp -> g'_call dest cont regenv exp
        (fun ys -> Call(l, ys)) ys
  | Save(x, y) -> assert false
(* ifのレジスタ割り当て *)
and g'_if dest cont regenv exp constr e1 e2 =
  let (e1', regenv1) = g dest cont regenv e1 in
  let (e2', regenv2) = g dest cont regenv e2 in
  let regenv' = (* 両方に共通のレジスタ変数だけ利用 *)
    S2.fold
      (fun x regenv' ->
        try
	  if is_reg x then regenv' else
          let r1 = M2.find x regenv1 in
          let r2 = M2.find x regenv2 in
          if r1 <> r2 then regenv' else
	  M2.add x r1 regenv'
        with Not_found -> regenv')
      (fv cont)
      M2.empty
  in
  (S2.fold
     (fun x e ->
       if x = dest || not (M2.mem x regenv) || M2.mem x regenv' then e else
       seq(Save(M2.find x regenv, x), e)) (* そうでない変数は分岐直前にセーブ *)
     (fv cont)
     (Ret(constr e1' e2')),
   regenv')
(* 関数呼び出しのレジスタ割り当て *)
and g'_call dest cont regenv exp constr ys =
  (S2.fold
     (fun x e ->
       if x = dest || not (M2.mem x regenv) then e else
       seq(Save(M2.find x regenv, x), e))
     (fv cont)
     (Ret(constr (List.map (fun y -> find y regenv) ys))),
   M2.empty)

(* 関数のレジスタ割り当て *)
let h { name = x; args = ys; body = e; stksiz = s } =
  let (i, arg_regs_rev, regenv) =
    List.fold_left
      (fun (i, arg_regs_rev, regenv) y ->
        let r = if y = reg_dummy then reg_dummy else regs.(i) in
        (i + 1, r :: arg_regs_rev, add y r regenv))
      (0, [], M2.empty)
      ys
  in
  let (e', regenv') = g regs.(0) (Ret(Mov regs.(0))) regenv e in
  { name = x; args = List.rev arg_regs_rev; body = e'; stksiz = s }

(* プログラム全体のレジスタ割り当て *)
let f fundefs =
  Format.eprintf "register allocation...@.";
  List.map h fundefs
