(* TreeAsm.t -> LinearAsm.t *)
open TreeAsm
open LinearAsm

let funs = ref M2.empty

(* 引数を所定のレジスタに移動する順番を決める *)
let rec shuffle xys =
  (* 移動する必要のないものを除く *)
  let xys = List.filter (fun (x,y) -> x <> y) xys in
  (* すぐ移動したら他の引数を潰してしまうものとそうでないものを分ける *)
  match List.partition (fun (_,y) -> List.mem_assoc y xys) xys with
  | [], [] -> []
  | (x,y) :: xys, [] -> (* 循環的に位置がずれているのでreg_tmpを経由 *)
      let xys_ = List.map
          (function (y_,z) when y = y_ -> (reg_tmp, z) | xy -> xy)
          xys
      in
      (y, reg_tmp) :: (x, y) :: shuffle xys_
  | xys, acyc -> acyc @ shuffle xys

let rec g dest c e =
  match e with
  | Ret exp -> g_ dest c exp
  | Let(x, exp, e) -> g_ x (g dest c e) exp
  | Forget(_, e) -> g dest c e
and g_ dest c exp =
  let g_if e1 e2 k =
    let c2 = g dest c e2 in
    let c1 = g dest (make (JUMP c.label) c2) e1 in
    make (k c2.label) c1
  in
  match exp with
  | Nop -> c
  | If(Eq,x,y,e1,e2) -> g_if e1 e2 (fun els -> BNE(x,y,els))
  | If(Lt,x,y,e1,e2) -> g_if e1 e2 (fun els -> BGE(x,y,els))
  | IfI(Eq,x,i,e1,e2) -> g_if e1 e2 (fun els -> BNEI(x,i,els))
  | IfI(Lt,x,i,e1,e2) -> g_if e1 e2 (fun els -> BGEI(x,i,els))
  | FIf(Eq,x,y,e1,e2) -> g_if e1 e2 (fun els -> BNE(x,y,els))
  | FIf(Lt,x,y,e1,e2) -> g_if e1 e2 (fun els -> FBGE(x,y,els))
  | St(x,y,I i) -> make (STORE(x, y, string_of_int i)) c
  | St(x,y,L l) -> make (STORE(x, y, l)) c
  | St(x,y,F a) -> assert false
  | Imm(I i) -> g_imm dest c (Int32.of_int i)
  | Imm(F a) -> g_imm dest c (f2b a)
  | Imm(L l) -> make (ADDI(dest, reg_zero, l)) c (* 手抜き *)
  | Rd -> make (GET dest) c
  | Call(x,ys) ->
      let rs =
        try (M2.find x !funs).args
        with Not_found -> regs_lst
      in
      let rec combine xs ys = match xs,ys with
      | x::xs, y::ys -> (x,y) :: combine xs ys
      | _ -> []
      in
      let yrs = shuffle (combine ys rs) in
      List.fold_right
        (fun (y,r) e -> g_ r e (Mov y))
        yrs
        (make (JAL x) c)
  | Mov x when x = reg_dummy || dest = reg_dummy || x = dest -> c
  | Mov x -> make (MOV(dest,x)) c
  | Op2(Add,x,y) -> make (ADD(dest,x,y)) c
  | Op2(Sub,x,y) -> make (SUB(dest,x,y)) c
  | Op2(And,x,y) -> make (AND(dest,x,y)) c
  | Op2(Or,x,y) -> make (OR(dest,x,y)) c
  | Op2(Xor,x,y) -> make (XOR(dest,x,y)) c
  | Op2(Xnor,x,y) -> make (XNOR(dest,x,y)) c
  | Op2(FAdd,x,y) -> make (FADD(dest,x,y)) c
  | Op2(FSub,x,y) -> make (FSUB(dest,x,y)) c
  | Op2(FMul,x,y) -> make (FMUL(dest,x,y)) c
  | Op2(Array,x,y) ->
      let b = make NOP c in
      let loop = make (STORE(y, reg_tmp, "0"))
          (make (ADDI(reg_tmp, reg_tmp, "1")) b)
      in
      b.exp <- BL(reg_tmp, reg_hp, loop.label);
      make (MOV(dest, reg_hp))
        (make (MOV(reg_tmp, reg_hp))
           (make (ADD(reg_hp, reg_hp, x)) loop))
  | Op1i(AddI,x,I i) -> make (ADDI(dest, x, string_of_int i)) c
  | Op1i(AddI,x,L l) -> make (ADDI(dest, x, l)) c
  | Op1i(Sll,x,I i) -> make (SLL(dest, x, i)) c
  | Op1i(Srr,x,I i) -> make (SRR(dest, x, i)) c
  | Op1i(Ld,x,I i) -> make (LOAD(dest, x, string_of_int i)) c
  | Op1i(Ld,x,L l) -> make (LOAD(dest, x, l)) c
  | Op1(FInv,x) -> make (FINV(dest, x)) c
  | Op1(FSqrt,x) -> make (FSQRT(dest, x)) c
  | Op1(Wr,x) -> make (PUT x) c
  | Op1i(_,_,F _) | Op1i(Sll,_,L _) | Op1i(Srr,_,L _) -> assert false
  | Save _ | Restore _ -> assert false
and g_imm dest c i32 =
  let lo = (Int32.to_int i32 lsl 15) asr 15 in
  let hi = Int32.to_int (Int32.shift_right i32 16) in
  let hi = hi + (lo asr 16) in
  let hi = (hi lsl 15) asr 15 in
  if hi = 0 then make (ADDI(dest, reg_zero, string_of_int lo)) c
  else make (ADDI(dest, reg_zero, string_of_int hi))
      (make (SLL(dest, dest, 16))
         (if lo = 0 then c
          else make (ADDI(dest, dest, string_of_int lo)) c))

let f xs =
  List.iter (fun func -> funs := M2.add func.name func !funs) xs;
  let ans = List.map
      (fun func ->
        let exit =
          if func.stksiz = 0 then make (JR reg_ra) end_node
          else make (LOAD(reg_ra, reg_sp, "0"))
              (make (ADDI(reg_sp, reg_sp, string_of_int func.stksiz))
                 (make (JR reg_ra) end_node))
        in
        let fbody = g regs.(0) exit func.body in
        let entry =
          if func.stksiz = 0 then fbody
          else make (ADDI(reg_sp, reg_sp, string_of_int (-func.stksiz)))
              (make (STORE(reg_ra, reg_sp, "0")) fbody)
        in
        { entry = entry; fbody = fbody; exit = exit;
          fname = func.name; stk = func.stksiz })
      xs
  in
  funs := M2.empty; ans
