(* 型推論 *)

open Id
open Shared
open FlatSyntax
open Type

exception Hoge of t * t

(* 型変数tv が型t の中に出現するかどうか *)
let rec occur tv t =
  match t with
  | Tvar({ contents = None } as tv2) -> tv == tv2
  | Tvar{ contents = Some t }
  | Tarray t -> occur tv t
  | Tfun(ts, t) -> occur tv t || List.exists (occur tv) ts
  | Ttuple ts -> List.exists (occur tv) ts
  | ForAll _ -> assert false
  | _ -> false

(* とっても間接的な参照を少しだけ間接的な参照に変えつつ、
   参照されているものを返す *)
let rec shorten t =
  match t with
  | Tvar({ contents = Some t2 } as tv) -> begin
      match t2 with
      | Tvar{ contents = Some t3 } ->
          tv := Some t3;
          shorten t3
      | _ -> shorten t2
    end
  | ForAll _ -> assert false
  | _ -> t

(* 多相型 -> 単相型 *)
let instantiate t =
  match t with
  | ForAll(tvs, t2) ->
      let m = List.map (fun tv -> tv, new_type ()) tvs in
      let rec iter t =
        match t with
        | Tvar({ contents = None } as tv) ->
            (try List.assq tv m with Not_found -> t)
        | Tvar{ contents = Some t } -> iter t
        | Tfun(ts, t) -> Tfun(List.map iter ts, iter t)
        | Tarray t -> Tarray(iter t)
        | Ttuple ts -> Ttuple(List.map iter ts)
        | ForAll _ -> assert false
        | _ -> t
      in
      iter t2
  | _ -> t

(* 出現する未定の型変数 *)
let tvs t =
  let rec iter acc t =
    match t with
    | Tvar({ contents = None } as tv) ->
        if List.memq tv acc then acc else tv::acc
    | Tvar{ contents = Some t } | Tarray t -> iter acc t
    | Ttuple ts -> List.fold_left iter acc ts
    | Tfun(ts, t) -> List.fold_left iter (iter acc t) ts
    | ForAll _ -> assert false
    | _ -> acc
  in
  match t with
  | ForAll(tvs, t) -> list_diffq (iter [] t) tvs
  | _ -> iter [] t

(* 単相型 -> 多相型 *)
let generalize exc t =
  assert (match t with ForAll _ -> false | _ -> true);
  let pol = list_diffq (tvs t) exc in
  if pol = [] then t
  else ForAll(pol, t)

(* value restriction *)
let rec generalizable e =
  match e with
  | V _ | Tuple _ -> true
  | If(_, e1, e2) -> generalizable e1 && generalizable e2
  | _ -> false

(* 単一化 *)
let rec unify t1 t2 =
  match shorten t1, shorten t2 with
  | Tvar tv1, Tvar tv2 when tv1 == tv2 -> ()
  | Tvar tv, t | t, Tvar tv ->
      assert (!tv = None);
      if occur tv t then failwith "recursive type"
      else tv := Some t
  | Tunit, Tunit | Tint, Tint | Tfloat, Tfloat | Tbool, Tbool -> ()
  | Tarray t1, Tarray t2 -> unify t1 t2
  | Ttuple t1s, Ttuple t2s -> List.iter2 unify t1s t2s
  | Tfun(t1s, t1), Tfun(t2s, t2) ->
      unify t1 t2;
      List.iter2 unify t1s t2s
  | _ -> raise (Hoge(t1,t2))(*failwith "unify"*)

(* コメント募集 *)
let tvs_in_env env =
  S.fold (fun x acc -> list_mergeq (tvs x.typ) acc) env []

(* 式を型推論 *)
let rec g env e =
  let find x =
    match x with
    | Unit -> Tunit
    | Int _ -> Tint
    | Float _ -> Tfloat
    | Bool _ -> Tbool
    | Var x -> instantiate x.typ
  in
  match e with
  | V x -> find x
  | Tuple xs -> Ttuple(List.map find xs)
  | Op(x, ys) ->
      let targs = List.map find ys in
      let tret = new_type () in
      unify (find x) (Tfun(targs, tret));
      tret
  | If(x, e1, e2) ->
      unify (find x) Tbool;
      let t = new_type () in
      unify (g env e1) t;
      unify (g env e2) t;
      t
  | Let([], e1, e2) ->
      unify (g env e1) Tunit;
      g env e2
  | Let([x], e1, e2) ->
      unify (g env e1) x.typ;
      if generalizable e1 then x.typ <- generalize (tvs_in_env env) x.typ;
      g (S.add x env) e2
  | Let(_::_::_ as xs, e1, e2) ->
      let t1 = g env e1 in
      let t_xs = Ttuple(List.map (fun x -> x.typ) xs) in
      unify t1 t_xs;
      if generalizable e1 then List.iter
        (fun x -> x.typ <- generalize (tvs_in_env env) x.typ)
        xs;
      g (S.add_list xs env) e2
  | LetRec(f, xs, e1, e2) ->
      let env2 = S.add f env in
      let env1 = S.add_list xs env2 in
      unify (Tfun(List.map (fun x -> x.typ) xs, g env1 e1)) f.typ;
      f.typ <- generalize (tvs_in_env env) f.typ;
      g env2 e2

(* コメント募集 *)
let rec deref_type t =
  match t with
  | Tunit | Tint | Tfloat -> t
  | Tbool -> Tint (* ひぎぃ *)
  | Tarray t -> Tarray(deref_type t)
  | Ttuple ts -> Ttuple(List.map deref_type ts)
  | Tfun(ts, t) -> Tfun(List.map deref_type ts, deref_type t)
  | Tvar{ contents = None } -> t
  | Tvar{ contents = Some t } -> deref_type t
  | ForAll(tvs, t) -> ForAll(tvs, deref_type t)

let rec deref_term e =
  match e with
  | V _ | Op _ | Tuple _ -> ()
  | If(_, e1, e2) -> deref_term e1; deref_term e2
  | Let(xs, e1, e2) ->
      List.iter (fun x -> x.typ <- deref_type x.typ) xs;
      deref_term e1; deref_term e2
  | LetRec(f, xs, e1, e2) -> f.typ <- deref_type f.typ;
      List.iter (fun x -> x.typ <- deref_type x.typ) xs;
      deref_term e1; deref_term e2

let f e =
  let env = S.add_list Builtin.fun_idents S.empty in
  unify (g env e) Tunit;
  deref_term e;
  e
